using System;
using System.ComponentModel;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Tutorial1.Terrain;

namespace TerrainTutorial
{
	/// <summary>
	/// This is the main type for your game
	/// </summary>
	public class Game1 : Microsoft.Xna.Framework.Game
	{
		#region Fields     
		private GraphicsDeviceManager _graphics;
		GraphicsDevice _device;

		bool _drawTerrain = true;

		RasterizerState _rsDefault = new RasterizerState();
		RasterizerState _rsWire = new RasterizerState();
		bool _isWire;
		KeyboardState _previousKeyboardState;
   
		Matrix _world = Matrix.Identity;
		DateTime _last = DateTime.Now; //For calculating FPS
		int _fps; //For calculating FPS

		Camera _camera;
		QuadTree _quadTree;

		#endregion

		public Game1()
		{
			_graphics = new GraphicsDeviceManager(this);
		}


		protected override void Initialize()
		{
			IsFixedTimeStep = false;
			_previousKeyboardState = Keyboard.GetState();
			base.Initialize();
		}

		/// <summary>
		/// LoadContent will be called once per game and is the place to load
		/// all of your content.
		/// </summary>
		protected override void LoadContent()
		{
			_rsDefault.CullMode = CullMode.CullCounterClockwiseFace;
			_rsDefault.FillMode = FillMode.Solid;

			_rsWire.CullMode = CullMode.CullCounterClockwiseFace;
			_rsWire.FillMode = FillMode.WireFrame;

			_graphics.PreferredBackBufferWidth = 1024;
			_graphics.PreferredBackBufferHeight = 800;
			_graphics.SynchronizeWithVerticalRetrace = false;
			_graphics.ApplyChanges();

			_device = _graphics.GraphicsDevice;

			Content.RootDirectory = "Content";

			Window.AllowUserResizing = true;
			IsMouseVisible = true;

			var heightMap = Content.Load<Texture2D>("hmSmall");

			_camera = new Camera(new Vector3(300, 1000, 300), new Vector3(300, 0, 290), _device, 7000f);

			_quadTree = new QuadTree(Vector3.Zero, heightMap, _camera.View, _camera.Projection, _device, 1);
			_quadTree.Effect.Texture = Content.Load<Texture2D>("jigsaw");
			_quadTree.MinimumDepth = 0;
			_quadTree.Cull = true;
			base.LoadContent();			
		}

		/// <summary>
		/// UnloadContent will be called once per game and is the place to unload
		/// all content.
		/// </summary>
		protected override void UnloadContent()
		{
			// TODO: Unload any non ContentManager content here
		}

		/// <summary>
		/// Allows the game to run logic such as updating the world,
		/// checking for collisions, gathering input, and playing audio.
		/// </summary>
		/// <param name="gameTime">Provides a snapshot of timing values.</param>
		protected override void Update(GameTime gameTime)
		{
			var currentKeyboardState = Keyboard.GetState();

			// Allows the game to exit
			if (currentKeyboardState.IsKeyDown(Keys.Escape))
				Exit();

			if (_previousKeyboardState.IsKeyDown(Keys.W) && !currentKeyboardState.IsKeyDown(Keys.W))
			{
				if (_isWire)
				{
					_device.RasterizerState = _rsDefault;
					_isWire = false;
				}
				else
				{
					_device.RasterizerState = _rsWire;
					_isWire = true;
				}
			}

			if (_previousKeyboardState.IsKeyDown(Keys.C) && !currentKeyboardState.IsKeyDown(Keys.C))
			{
				_quadTree.Cull = !_quadTree.Cull;
			}	

			if (_previousKeyboardState.IsKeyDown(Keys.D) && !currentKeyboardState.IsKeyDown(Keys.D))
			{
				_drawTerrain = !_drawTerrain;
			}

			if (currentKeyboardState[Keys.Up] == KeyState.Down)
				_camera.Move(new Vector3(0, 0, -0.05f));
			if (currentKeyboardState[Keys.Down] == KeyState.Down)
				_camera.Move(new Vector3(0, 0, 0.05f));
			if (currentKeyboardState[Keys.Left] == KeyState.Down)
				_camera.Move(new Vector3(-0.05f, 0, 0));
			if (currentKeyboardState[Keys.Right] == KeyState.Down)
				_camera.Move(new Vector3(0.05f, 0, 0));

			if (currentKeyboardState[Keys.PageDown] == KeyState.Down)
				_camera.Move(new Vector3(0, -0.05f, 0));
			if (currentKeyboardState[Keys.PageUp] == KeyState.Down)
				_camera.Move(new Vector3(0, 0.05f, 0));

			if (currentKeyboardState[Keys.A] == KeyState.Down)
				_camera.Pitch(new Vector3(0, 0, -0.1f));
			if (currentKeyboardState[Keys.Z] == KeyState.Down)
				_camera.Pitch(new Vector3(0, 0, 0.1f));

			_previousKeyboardState = currentKeyboardState;
			_quadTree.View = _camera.View;
			_quadTree.Projection = _camera.Projection;
			_quadTree.CameraPosition = _camera.Position;
			_quadTree.Update(gameTime);


			if ((DateTime.Now - _last).TotalMilliseconds >= 1000)
			{
				var title = string.Concat(_fps, " fps | ", "Vertices: ", _quadTree.Vertices.Vertices.Length, " | Tris: ", _quadTree.IndexCount / 3, " | CULLING: " + _quadTree.Cull);
				Window.Title = title;
				_fps = 0;
				_last = DateTime.Now;
			}
			else
			{
				_fps++;
			}

			base.Update(gameTime);
		}

		/// <summary>
		/// This is called when the game should draw itself.
		/// </summary>
		/// <param name="gameTime">Provides a snapshot of timing values.</param>
		protected override void Draw(GameTime gameTime)
		{
			_device.Clear(Color.Black);

			if(_drawTerrain)
				_quadTree.Draw(gameTime);

			base.Draw(gameTime);
		}
	}
}
